"""
hubspot companies api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import prettify, get_log
from typing import List, Dict, Optional, Union


COMPANIES_API_VERSION = "2"


class CompaniesClient(BaseClient):
    """
    hubspot3 Companies client
    :see: https://developers.hubspot.com/docs/methods/companies/companies-overview
    """

    def __init__(self, *args, **kwargs):
        super(CompaniesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.companies")

    def _get_path(self, subpath: str) -> str:
        """get the full api url for the given subpath on this client"""
        return "companies/v{}/{}".format(
            self.options.get("version") or COMPANIES_API_VERSION, subpath
        )

    def create(self, data: Dict = None, **options) -> Dict:
        """create a new company"""
        data = data or {}
        return self._call("companies/", data=data, method="POST", **options)

    def update(self, company_id: str, data: Dict = None, **options) -> Dict:
        """update the given company with data"""
        data = data or {}
        return self._call(
            "companies/{}".format(company_id), data=data, method="PUT", **options
        )

    def delete(self, company_id: str, **options) -> Dict:
        """delete a company"""
        return self._call("companies/{}".format(company_id), method="DELETE", **options)

    def delete_all(self, **options):
        """
        Delete all the companies. Please use it carefully.
        """
        for company in self.get_all(**options):
            self.delete(company["id"])

    def get(self, company_id: str, **options) -> Dict:
        """get a single company by it's ID"""
        return self._call("companies/{}".format(company_id), method="GET", **options)

    def search_domain(
        self, domain: str, limit: int = 1, extra_properties: Dict = None, **options
    ) -> Dict:
        """searches for companies by domain name. limit is max'd at 100"""
        # default properties to fetch
        properties = [
            "domain",
            "createdate",
            "name",
            "hs_lastmodifieddate",
            "hubspot_owner_id",
        ]

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties += extra_properties
            if isinstance(extra_properties, str):
                properties.append(extra_properties)

        return self._call(
            "domains/{}/companies".format(domain),
            method="POST",
            data={"limit": limit, "requestOptions": {"properties": properties}},
            **options
        )

    def get_all(
        self, extra_properties: Union[str, List] = None, **options
    ) -> Optional[List]:
        """get all companies, including extra properties if they are passed in"""
        finished = False
        output = []
        offset = 0
        query_limit = 250  # Max value according to docs

        # default properties to fetch
        properties = [
            "name",
            "description",
            "address",
            "address2",
            "city",
            "state",
            "story",
            "hubspot_owner_id",
        ]

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties += extra_properties
            if isinstance(extra_properties, str):
                properties.append(extra_properties)

        while not finished:
            batch = self._call(
                "companies/paged",
                method="GET",
                doseq=True,
                params={
                    "limit": query_limit,
                    "offset": offset,
                    "properties": properties,
                },
                **options
            )
            output.extend(
                [
                    prettify(company, id_key="companyId")
                    for company in batch["companies"]
                    if not company["isDeleted"]
                ]
            )
            finished = not batch["has-more"]
            offset = batch["offset"]

        return output

    def _get_recent(self, recency_type: str, **options) -> Optional[List]:
        """
        Returns either list of recently modified companies or recently created companies,
        depending on recency_type passed in. Both API endpoints take identical parameters
        and return identical formats, they differ only in the URLs
        (companies/recent/created or companies/recent/modified)
        :see: https://developers.hubspot.com/docs/methods/companies/get_companies_modified
        :see: https://developers.hubspot.com/docs/methods/companies/get_companies_created
        """
        finished = False
        output = []
        offset = 0
        query_limit = 250  # Max value according to docs

        while not finished:
            batch = self._call(
                "companies/recent/{}".format(recency_type),
                method="GET",
                doseq=True,
                params={"count": query_limit, "offset": offset},
                **options
            )
            output.extend(
                [
                    prettify(company, id_key="companyId")
                    for company in batch["results"]
                    if not company["isDeleted"]
                ]
            )
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def get_recently_modified(self, **options) -> Optional[List]:
        return self._get_recent("modified", **options)

    def get_recently_created(self, **options) -> Optional[List]:
        return self._get_recent("created", **options)

    def get_contacts_at_a_company(self, company_id: str, **options) -> Optional[List]:
        """
        Returns all of the contacts who have an associatedcompanyid contact property of
        `company_id`.

        :see: https://developers.hubspot.com/docs/methods/companies/get_company_contacts
        """
        return self._call(
            "companies/{}/contacts".format(company_id), method="GET", **options
        )



"""
hubspot cms_layout api client
"""
from hubspot3.base import BaseClient


LAYOUTS_API_VERSION = "2"


class CMSLayoutsClient(BaseClient):
    """
    provides a client for accessing hubspot layout info
    """

    def _get_path(self, subpath: str) -> str:
        return "content/api/v{}/layouts/{}".format(LAYOUTS_API_VERSION, subpath)

    def get_layouts(self, **options):
        return self._call("", **options)

    def get_layout_info(self, layout_id: str, **options):
        return self._call("{}".format(layout_id), **options)

    def get_layout_buffer(self, layout_id: str, **options):
        return self._call("{}/buffer".format(layout_id), **options)

    def get_layout_has_buffered_changes(self, layout_id: str, **options):
        return self._call("{}/has_buffered_changes".format(layout_id), **options)

    def get_layout_versions(self, layout_id: str, **options):
        return self._call("{}/versions".format(layout_id), **options)

    def get_layout_version_info(self, layout_id: str, version_id: str, **options):
        return self._call("{}/versions/{}".format(layout_id, version_id), **options)



"""
hubspot deals api
"""
import urllib.parse
from hubspot3.base import BaseClient
from hubspot3.utils import prettify, get_log
from typing import Union


DEALS_API_VERSION = "1"


class DealsClient(BaseClient):
    """
    The hubspot3 Deals client uses the _make_request method to call the API
    for data.  It returns a python object translated from the json returned
    """

    class Recency:
        """recency type enum"""

        CREATED = "created"
        MODIFIED = "modified"

    def __init__(self, *args, **kwargs):
        super(DealsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.deals")

    def _get_path(self, subpath):
        return "deals/v{}/{}".format(
            self.options.get("version") or DEALS_API_VERSION, subpath
        )

    def get(self, deal_id: str, **options):
        return self._call("deal/{}".format(deal_id), method="GET", **options)

    def create(self, data: dict = None, **options):
        """
        create deal api call
        :see: https://developers.hubspot.com/docs/methods/deals/create_deal
        """
        data = data or {}
        return self._call("deal/", data=data, method="POST", **options)

    def update(self, deal_id: str, data: dict = None, **options):
        data = data or {}
        return self._call("deal/{}".format(deal_id), data=data, method="PUT", **options)

    def associate(self, deal_id, object_type, object_ids, **options):
        # Encoding the query string here since HubSpot is expecting the "id" parameter to be
        # repeated for each object ID, which is not a standard practice and won't work otherwise.
        object_ids = [("id", object_id) for object_id in object_ids]
        query = urllib.parse.urlencode(object_ids)

        return self._call(
            "deal/{}/associations/{}".format(deal_id, object_type),
            method="PUT",
            query=query,
            **options
        )

    def get_all(
        self,
        offset: int = 0,
        extra_properties: Union[list, str] = None,
        limit: int = -1,
        **options
    ):
        """
        get all deals in the hubspot account.
        extra_properties: a list used to extend the properties fetched
        :see: https://developers.hubspot.com/docs/methods/deals/get-all-deals
        """
        finished = False
        output = []
        query_limit = 250  # Max value according to docs
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit

        # default properties to fetch
        properties = [
            "associations",
            "dealname",
            "dealstage",
            "pipeline",
            "hubspot_owner_id",
            "description",
            "closedate",
            "amount",
            "dealtype",
            "createdate",
        ]

        # append extras if they exist
        if extra_properties:
            if isinstance(extra_properties, list):
                properties += extra_properties
            if isinstance(extra_properties, str):
                properties.append(extra_properties)

        while not finished:
            batch = self._call(
                "deal/paged",
                method="GET",
                params={
                    "limit": query_limit,
                    "offset": offset,
                    "properties": properties,
                    "includeAssociations": True,
                },
                doseq=True,
                **options
            )
            output.extend(
                [
                    prettify(deal, id_key="dealId")
                    for deal in batch["deals"]
                    if not deal["isDeleted"]
                ]
            )
            finished = not batch["hasMore"] or (limited and len(output) >= limit)
            offset = batch["offset"]

        return output if not limited else output[:limit]

    def _get_recent(
        self,
        recency_type: str,
        limit: int = 100,
        offset: int = 0,
        since: int = None,
        include_versions: bool = False,
        **options
    ):
        """
        returns a list of either recently created or recently modified deals

        :param since: unix formatted timestamp in milliseconds
        """
        finished = False
        output = []
        query_limit = 100  # max according to the docs
        limited = limit > 0
        if limited and limit < query_limit:
            query_limit = limit

        while not finished:
            params = {
                "count": query_limit,
                "offset": offset,
                "includePropertyVersions": include_versions,
            }
            if since:
                params["since"] = since
            batch = self._call(
                "deal/recent/{}".format(recency_type),
                method="GET",
                params=params,
                doseq=True,
                **options
            )
            output.extend(
                [
                    prettify(deal, id_key="dealId")
                    for deal in batch["results"]
                    if not deal["isDeleted"]
                ]
            )
            finished = not batch["hasMore"] or len(output) >= limit
            offset = batch["offset"]

        return output[:limit]

    def get_recently_created(
        self,
        limit: int = 100,
        offset: int = 0,
        since: int = None,
        include_versions: bool = False,
        **options
    ):
        """
        get recently created deals
        up to the last 30 days or the 10k most recently created records

        since: must be a UNIX formatted timestamp in milliseconds
        """
        return self._get_recent(
            DealsClient.Recency.CREATED,
            limit=limit,
            offset=offset,
            since=since,
            include_versions=include_versions,
            **options
        )

    def get_recently_modified(
        self,
        limit: int = 100,
        offset: int = 0,
        since: int = None,
        include_versions: bool = False,
        **options
    ):
        """
        get recently modified deals
        up to the last 30 days or the 10k most recently modified records

        since: must be a UNIX formatted timestamp in milliseconds
        """
        return self._get_recent(
            DealsClient.Recency.MODIFIED,
            limit=limit,
            offset=offset,
            since=since,
            include_versions=include_versions,
            **options
        )



"""
hubspot crm_pipelines api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


CRM_PIPELINES_API_VERSION = "1"


class PipelinesClient(BaseClient):
    """
    The hubspot3 Pipelines client uses the _make_request method to call the API
    for data.  It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        super(PipelinesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.crm-pipelines")

    def _get_path(self, subpath):
        return "crm-pipelines/v{}/{}".format(
            self.options.get("version") or CRM_PIPELINES_API_VERSION, subpath
        )

    def create(self, object_type, data=None, **options):
        data = data or {}
        return self._call(
            "pipeline/{}".format(object_type), data=data, method="POST", **options
        )

    def update(self, object_type, key, data=None, **options):
        data = data or {}
        return self._call(
            "pipeline/{}/{}".format(object_type, key),
            data=data,
            method="PUT",
            **options
        )

    def get_all(self, object_type="deals", offset=0, extra_properties=None, **options):
        """
        get all crm_pipelines in the hubspot account with given object_type.
        object_type (default='deals'): must be 'deals' or 'tickets'
        extra_properties: a list used to extend the properties fetched
        """
        output = []

        output = self._call("pipelines/{}".format(object_type), method="GET", **options)

        return output["results"]

    def get_deals_pipeline_by_id(self, pipeline_id: str):
        """
        Retrieve a deals pipeline by its id.

        Notes: At the moment, it is impossible to retrieve a pipeline by id directly through the
        hubspot API. We have to fetch all the pipelines of type 'DEAL' and then looks for a
        pipeline with the given `pipeline_id`.

        Returns
        -------
        dict
            The pipeline as returned by the Hubspot API.
            See the docstring of the `get_all_for_deals` method below to see an output example.

            None could be returned if no pipeline is matching with the given `pipeline_id`.
        """
        pipelines = self.get_all()
        for pipeline in pipelines:
            if pipeline.get("pipelineId") == pipeline_id:
                return pipeline
        return None



"""
hubspot email subscription api
"""
from typing import Dict, Iterable, Mapping
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


EMAIL_SUBSCRIPTION_API_VERSION = "1"


class EmailSubscriptionClient(BaseClient):
    """
    The hubspot3 Email Subscription client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    class OptState:
        """opt state enum"""

        OPT_IN = "OPT_IN"
        OPT_OUT = "OPT_OUT"
        NOT_OPTED = "NOT_OPTED"

    class LegalBasis:
        """legal basis enum"""

        CONSENT_WITH_NOTICE = "CONSENT_WITH_NOTICE"
        LEGITIMATE_INTEREST_PQL = "LEGITIMATE_INTEREST_PQL"
        LEGITIMATE_INTEREST_CLIENT = "LEGITIMATE_INTEREST_CLIENT"
        NON_GDPR = "NON_GDPR"
        PERFORMANCE_OF_CONTRACT = "PERFORMANCE_OF_CONTRACT"

    def __init__(self, *args, **kwargs):
        """initialize an email subscription client"""
        super(EmailSubscriptionClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.email_subscription")

    def _get_path(self, subpath):
        return "email/public/v{}/subscriptions/{}".format(
            EMAIL_SUBSCRIPTION_API_VERSION, subpath
        )

    def get_status(self, email: str, portal_id: int = None, **options) -> Dict:
        """
        Retrieve the email subscription status for the given email address.
        :see: https://developers.hubspot.com/docs/methods/email/get_status
        """
        params = {}
        if portal_id is not None:
            params["portalId"] = portal_id
        return self._call(email, method="GET", params=params, **options)

    def update_status(self, email: str, data: Mapping, **options) -> None:
        """
        Update the email subscription status for the given email address using the given raw data.
        :see: https://developers.hubspot.com/docs/methods/email/update_status
        """
        self._call(email, method="PUT", data=dict(data), **options)

    def update_subscriptions(
        self,
        email: str,
        subscriptions: Iterable,
        portal_legal_basis: str = None,
        portal_legal_basis_explanation: str = None,
        **options
    ) -> None:
        """
        Convenience method to update the individual email subscriptions for a given email address.
        :see: https://developers.hubspot.com/docs/methods/email/update_status
        """
        data = {"subscriptionStatuses": list(subscriptions)}
        if portal_legal_basis:
            data["portalSubscriptionLegalBasis"] = portal_legal_basis
        if portal_legal_basis_explanation:
            data[
                "portalSubscriptionLegalBasisExplanation"
            ] = portal_legal_basis_explanation
        self.update_status(email, data, **options)

    def unsubscribe_permanently(self, email: str, **options) -> None:
        """
        Convenience method to unsubscribe an email address from all email subscriptions
        permanently, i.e. this email address cannot be added opted into any email subscription
        anymore.
        :see: https://developers.hubspot.com/docs/methods/email/update_status
        """
        self.update_status(email, {"unsubscribeFromAll": True}, **options)



"""
hubspot properties api
"""
from hubspot3.base import BaseClient
from hubspot3.globals import (
    OBJECT_TYPE_COMPANIES,
    OBJECT_TYPE_CONTACTS,
    OBJECT_TYPE_DEALS,
    OBJECT_TYPE_LINE_ITEMS,
    OBJECT_TYPE_PRODUCTS,
    VALID_PROPERTY_DATA_TYPES,
    VALID_PROPERTY_WIDGET_TYPES,
)
from hubspot3.utils import get_log

PROPERTIES_API_VERSION = {
    OBJECT_TYPE_COMPANIES: "1",
    OBJECT_TYPE_CONTACTS: "1",
    OBJECT_TYPE_DEALS: "1",
    OBJECT_TYPE_LINE_ITEMS: "2",
    OBJECT_TYPE_PRODUCTS: "2",
}


class PropertiesClient(BaseClient):
    """
    The hubspot3 Properties client uses the _make_request method to call the
    API for data.  It returns a python object translated from the json returned
    """

    # Used to store a current object_type. This will be used in `_get_path` to generate the
    # appropriate path to use.
    _object_type = None

    def __init__(self, *args, **kwargs):
        super(PropertiesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.properties")

    def _get_path(self, subpath):
        return "properties/v{}/{}/properties/{}".format(
            PROPERTIES_API_VERSION[self._object_type], self._object_type, subpath
        )

    def create(
        self,
        object_type,
        code,
        label,
        description,
        group_code,
        data_type,
        widget_type,
        extra_params=None,
    ):
        """
        Create a new custom property on hubspot.
        """

        if data_type not in VALID_PROPERTY_DATA_TYPES:
            raise ValueError(
                "Invalid data type for property. Valid data types are: {}".format(
                    VALID_PROPERTY_DATA_TYPES
                )
            )

        if widget_type not in VALID_PROPERTY_WIDGET_TYPES:
            raise ValueError(
                "Invalid widget type for property. Valid widget types are: {}".format(
                    VALID_PROPERTY_WIDGET_TYPES
                )
            )

        extra_params = extra_params or {}

        # Save the current object type.
        self._object_type = object_type

        return self._call(
            "",
            method="POST",
            data={
                "name": code,
                "label": label,
                "description": description,
                "groupName": group_code,
                "type": data_type,
                "fieldType": widget_type,
                **extra_params,
            },
        )

    def get_all(self, object_type):
        """Retrieve all the custom properties."""

        # Save the current object type.
        self._object_type = object_type

        return self._call(
            "", method="GET", params={"properties": ["name", "label", "description"]}
        )

    def delete(self, object_type, code):
        """Delete a custom property."""

        # Save the current object type.
        self._object_type = object_type

        return self._call("named/{}".format(code), method="DELETE")

    def delete_all(self, object_type):
        """Delete all the custom properties. Please use it carefully."""
        props_data = self.get_all(object_type)

        for prop_data in props_data:
            if not prop_data["hubspotDefined"]:
                self.delete(object_type, prop_data["name"])



"""
hubspot keywords api
"""
from hubspot3.base import BaseClient

KEYWORDS_API_VERSION = "v1"


class KeywordsClient(BaseClient):
    """allows access to the keywords api"""

    def _get_path(self, subpath):
        return "keywords/{}/{}".format(KEYWORDS_API_VERSION, subpath)

    # Contains both list of keywords and metadata
    def get_keywords_info(self, **options):
        return self._call("keywords", **options)

    # *Only* returns the list of keywords, does not include additional metadata
    def get_keywords(self, **options):
        return self._call("keywords", **options)["keywords"]

    def get_keyword(self, keyword_guid, **options):
        return self._call("keywords/{}".format(keyword_guid, **options))

    def add_keyword(self, keyword, **options):
        return self._call(
            "keywords", data=dict(keyword=str(keyword)), method="PUT", **options
        )

    def add_keywords(self, keywords, **options):
        data = []
        for keyword in keywords:
            if keyword != "":
                if isinstance(keyword, dict):
                    data.append(keyword)
                elif isinstance(keyword, str):
                    data.append(dict(keyword=str(keyword)))
        return self._call("keywords", data=data, method="PUT", **options)["keywords"]

    def delete_keyword(self, keyword_guid, **options):
        return self._call(
            "keywords/{}".format(keyword_guid), method="DELETE", **options
        )



"""
hubspot blog api client
"""
import json
from hubspot3.base import BaseClient
from typing import Dict


BLOG_API_VERSION = "2"
COMMENTS_API_VERSION = "3"
TOPICS_API_VERSION = "3"


class BlogClient(BaseClient):
    """
    provides a client for accessing hubspot blog info
    """

    def _get_path(self, subpath: str) -> str:
        return "content/api/v{}/{}".format(BLOG_API_VERSION, subpath)

    def get_blogs(self, **options):
        return self._call("blogs", **options)

    def get_blog_info(self, blog_guid, **options):
        return self._call("blogs/{}".format(blog_guid), **options)

    def get_posts(self, blog_guid: str, **options) -> Dict:
        if "params" not in options:
            options["params"] = {}
        options["params"].update({"content_group_id": blog_guid})
        return self._call("blog-posts", **options)

    def get_draft_posts(self, blog_guid: str, **options) -> Dict:
        if "params" not in options:
            options["params"] = {}
        options["params"].update({"content_group_id": blog_guid, "state": "DRAFT"})
        return self._call("blog-posts", **options)

    def get_published_posts(self, blog_guid: str, **options) -> Dict:
        if "params" not in options:
            options["params"] = {}
        options["params"].update({"content_group_id": blog_guid, "state": "PUBLISHED"})
        return self._call("blog-posts", **options)

    # Spelled wrong but left for compat
    def get_pulished_posts(self, blog_guid: str, **options) -> Dict:
        return self.get_published_posts(blog_guid, **options)

    def get_post(self, post_guid: str, **options) -> Dict:
        return self._call("blog-posts/{}".format(post_guid), **options)

    def create_post(
        self, blog_guid, author_id, title, summary, content, meta_desc, **options
    ):
        post = json.dumps(
            dict(
                content_group_id=blog_guid,
                name=title,
                blog_author_id=author_id,
                post_summary=summary,
                post_body=content,
                meta_description=meta_desc,
            )
        )
        raw_response = self._call(
            "blog-posts",
            data=post,
            method="POST",
            content_type="application/json",
            raw_output=True,
            **options
        )
        return raw_response

    def update_post(
        self,
        post_guid,
        title=None,
        summary=None,
        content=None,
        meta_desc=None,
        **options
    ):
        update_param_translation = dict(
            title="name",
            summary="post_summary",
            content="post_body",
            meta_desc="meta_description",
        )
        post_dict = dict(
            [
                (k, locals()[p])
                for p, k in update_param_translation.items()
                if locals().get(p)
            ]
        )
        post = json.dumps(post_dict)
        raw_response = self._call(
            "blog-posts/{}".format(post_guid),
            data=post,
            method="PUT",
            content_type="application/json",
            raw_output=True,
            **options
        )
        return raw_response

    def publish_post(self, post_guid: str, **options) -> Dict:
        post = json.dumps(dict(action="schedule-publish"))
        raw_response = self._call(
            "blog-posts/{}/publish-action".format(post_guid),
            data=post,
            method="PUT",
            content_type="application/json",
            raw_output=True,
            **options
        )
        return raw_response


class BlogCommentsClient(BaseClient):
    """
    provides a client for accessing hubspot comments info
    """

    def _get_path(self, subpath: str) -> str:
        return "comments/v{}/{}".format(COMMENTS_API_VERSION, subpath)

    def get_comments(self, **options):
        return self._call("comments", **options)

    def get_post_comments(self, post_guid: str, **options) -> Dict:
        if "params" not in options:
            options["params"] = {}
        options["params"].update({"contentId": post_guid})
        return self._call("comments", **options)

    def get_comment(self, comment_guid: str, **options) -> Dict:
        return self._call("comments/{}".format(comment_guid), **options)

    def create_comment(
        self,
        blog_guid,
        post_guid,
        author_name,
        author_email,
        author_uri,
        content,
        **options
    ):
        post = dict(
            collectionId=blog_guid,
            contentId=post_guid,
            userName=author_name,
            userEmail=author_email,
            userUrl=author_uri,
            comment=content,
        )
        raw_response = self._call(
            "comments",
            data=post,
            method="POST",
            content_type="application/json",
            raw_output=True,
            **options
        )
        return raw_response


class BlogTopicsClient(BaseClient):
    """
    provides a client for accessing hubspot blog topics info
    """

    def _get_path(self, subpath: str) -> str:
        return "blogs/v{}/{}".format(TOPICS_API_VERSION, subpath)

    def get_topics(self, **options):
        return self._call("topics", **options)



"""
hubspot engagements api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


ENGAGEMENTS_API_VERSION = "1"


class EngagementsClient(BaseClient):
    """
    The hubspot3 Engagements client uses the _make_request method to call the API
    for data.  It returns a python object translated from the json returned
    """

    def __init__(self, *args, **kwargs):
        super(EngagementsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.engagements")

    def _get_path(self, subpath):
        return "engagements/v{}/{}".format(
            self.options.get("version") or ENGAGEMENTS_API_VERSION, subpath
        )

    def get(self, engagement_id, **options):
        """Get a HubSpot engagement."""
        return self._call(
            "engagements/{}".format(engagement_id), method="GET", **options
        )

    def get_associated(self, object_type, object_id, **options):
        """
        get all engagements associated with the given object
        :param object_type: type of object to get associations on [CONTACT, COMPANY, DEAL]
        :param object_id: ID of the object to get associations on
        """
        finished = False
        output = []
        query_limit = 100  # Max value according to docs
        offset = 0
        while not finished:
            print(offset)
            batch = self._call(
                "engagements/associated/{}/{}/paged".format(object_type, object_id),
                method="GET",
                params={"limit": query_limit, "offset": offset},
                **options
            )
            print(len(batch["results"]))
            output.extend(batch["results"])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def create(self, data=None, **options):
        data = data or {}
        return self._call("engagements", data=data, method="POST", **options)

    def update(self, key, data=None, **options):
        data = data or {}
        return self._call(
            "engagements/{}".format(key), data=data, method="PUT", **options
        )

    def get_all(self, **options):
        """get all engagements"""
        finished = False
        output = []
        query_limit = 250  # Max value according to docs
        offset = 0
        while not finished:
            batch = self._call(
                "engagements/paged",
                method="GET",
                params={"limit": query_limit, "offset": offset},
                **options
            )
            output.extend(batch["results"])
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output



"""
hubspot lines api
"""
from hubspot3.crm_associations import CRMAssociationsClient
from hubspot3.base import BaseClient
from hubspot3.utils import get_log


LINES_API_VERSION = "1"


class LinesClient(BaseClient):
    """
    Line Items API endpoint
    THIS API ENDPOINT IS ONLY A PREVIEW AND IS SUBJECT TO CHANGE
    :see: https://developers.hubspot.com/docs/methods/line-items/line-items-overview
    """

    def __init__(self, *args, **kwargs):
        super(LinesClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.lines")

    def create(self, data=None, **options):
        return self._call("", data=data, method="POST", **options)

    def get(self, line_id, **options):
        """Retrieve a line by its id."""
        return self._call("{}".format(line_id), **options)

    def link_line_item_to_deal(self, line_item_id, deal_id):
        """Link a line item to a deal."""
        associations_client = CRMAssociationsClient(**self.credentials)
        return associations_client.link_line_item_to_deal(line_item_id, deal_id)

    def _get_path(self, subpath: str):
        return "crm-objects/v{}/objects/line_items/{}".format(
            self.options.get("version") or LINES_API_VERSION, subpath
        )



"""
hubspot3 module
"""
from datetime import datetime, timedelta
from hubspot3.error import HubspotBadConfig, HubspotNoConfig
from typing import Any


class Hubspot3UsageLimits:
    """a nicer wrapper for the usage limit data"""

    class FetchStatus:
        """fetch status enum"""

        NONE = ""
        SUCCESS = "SUCCESS"
        CACHED = "CACHED"

    def __init__(
        self,
        collected_at: datetime = datetime.fromtimestamp(0),
        current_usage: int = 0,
        fetch_status: str = FetchStatus.NONE,
        resets_at: datetime = datetime.fromtimestamp(0),
        usage_limit: int = 0,
    ) -> None:
        """usage limits constructor"""
        self.collected_at = collected_at
        self.current_usage = current_usage
        self.fetch_status = fetch_status
        self.resets_at = resets_at
        self.usage_limit = usage_limit

    def __str__(self) -> str:
        """string representation"""
        return "<Hubspot3UsageLimits: {}/{} ({}%) [reset in {}s, cached for {}s]>".format(
            self.current_usage,
            self.usage_limit,
            self.calls_used,
            self.until_reset,
            self.until_cache_expire,
        )

    def __repr__(self) -> str:
        """repr, just use __str__"""
        return self.__str__()

    @property
    def until_reset(self) -> int:
        """returns the number of seconds until limit reset"""
        return int((self.resets_at - datetime.now()).total_seconds())

    @property
    def until_cache_expire(self):
        """returns the number of seconds until the cache expires, and we need an update"""
        return int(
            (self.collected_at + timedelta(minutes=5) - datetime.now()).total_seconds()
        )

    @property
    def calls_remaining(self) -> int:
        """returns the number of calls remaining in your usage limits for the day"""
        return self.usage_limit - self.current_usage

    @property
    def calls_used(self) -> float:
        """returns the percentage of all calls used"""
        return self.current_usage / self.usage_limit

    @property
    def need_update(self):
        """
        true if the data is 'old'.
        data returned by the daily limit endpoint is cached on their end for 5 minutes.
        :see: https://developers.hubspot.com/docs/methods/check-daily-api-usage
        """
        five_minutes_ago = datetime.now() - timedelta(minutes=5)
        if self.collected_at and self.collected_at > five_minutes_ago:
            return False
        return True


class Hubspot3:
    """hubspot3 full client class"""

    def __init__(
        self,
        api_key: str = None,
        access_token: str = None,
        refresh_token: str = None,
        client_id: str = None,
        client_secret: str = None,
        timeout: int = 10,
        api_base: str = "api.hubapi.com",
        debug: bool = False,
        disable_auth: bool = False,
        **extra_options: Any
    ) -> None:
        """full client constructor"""
        self.api_key = api_key
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.client_id = client_id
        self.client_secret = client_secret
        if not disable_auth:
            if self.api_key and self.access_token:
                raise HubspotBadConfig("Cannot use both api_key and access_token.")
            if not (self.api_key or self.access_token or self.refresh_token):
                raise HubspotNoConfig("Missing required credentials.")
        self.auth = {
            "access_token": self.access_token,
            "api_key": self.api_key,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
        }
        self.options = {
            "api_base": api_base,
            "debug": debug,
            "disable_auth": disable_auth,
            "timeout": timeout,
        }
        self.options.update(extra_options)

        # rate limiting related stuff
        self._usage_limits = Hubspot3UsageLimits()

    @property
    def _base(self):
        """returns a hubspot3 base client"""
        from hubspot3.base import BaseClient

        return BaseClient(**self.auth, **self.options)

    @property
    def blog(self):
        """returns a hubspot3 blog client"""
        from hubspot3.blog import BlogClient

        return BlogClient(**self.auth, **self.options)

    @property
    def blog_comments(self):
        """returns a hubspot3 blog comments client"""
        from hubspot3.blog import BlogCommentsClient

        return BlogCommentsClient(**self.auth, **self.options)

    @property
    def blog_topics(self):
        """returns a hubspot3 blog topics client"""
        from hubspot3.blog import BlogTopicsClient

        return BlogTopicsClient(**self.auth, **self.options)

    @property
    def broadcast(self):
        """returns a hubspot3 broadcast client"""
        from hubspot3.broadcast import BroadcastClient

        return BroadcastClient(**self.auth, **self.options)

    @property
    def cms_layouts(self):
        """returns a hubspot3 layouts client"""
        from hubspot3.cms_layouts import CMSLayoutsClient

        return CMSLayoutsClient(**self.auth, **self.options)

    @property
    def cms_templates(self):
        """returns a hubspot3 templates client"""
        from hubspot3.cms_templates import CMSTemplatesClient

        return CMSTemplatesClient(**self.auth, **self.options)

    @property
    def companies(self):
        """returns a hubspot3 companies client"""
        from hubspot3.companies import CompaniesClient

        return CompaniesClient(**self.auth, **self.options)

    @property
    def contact_lists(self):
        """returns a hubspot3 contact_lists client"""
        from hubspot3.contact_lists import ContactListsClient

        return ContactListsClient(**self.auth, **self.options)

    @property
    def contacts(self):
        """returns a hubspot3 contacts client"""
        from hubspot3.contacts import ContactsClient

        return ContactsClient(**self.auth, **self.options)

    @property
    def crm_associations(self):
        """returns a hubspot3 crm_associations client"""
        from hubspot3.crm_associations import CRMAssociationsClient

        return CRMAssociationsClient(**self.auth, **self.options)

    @property
    def crm_pipelines(self):
        """returns a hubspot3 crm_pipelines client"""
        from hubspot3.crm_pipelines import PipelinesClient

        return PipelinesClient(**self.auth, **self.options)

    @property
    def deals(self):
        """returns a hubspot3 deals client"""
        from hubspot3.deals import DealsClient

        return DealsClient(**self.auth, **self.options)

    @property
    def ecommerce_bridge(self):
        """returns a hubspot3 ecommerce bridge client"""
        from hubspot3.ecommerce_bridge import EcommerceBridgeClient

        return EcommerceBridgeClient(**self.auth, **self.options)

    @property
    def email_subscription(self):
        """returns a hubspot3 email subscription client"""
        from hubspot3.email_subscription import EmailSubscriptionClient

        return EmailSubscriptionClient(**self.auth, **self.options)

    @property
    def engagements(self):
        """returns a hubspot3 engagements client"""
        from hubspot3.engagements import EngagementsClient

        return EngagementsClient(**self.auth, **self.options)

    @property
    def form_submissions(self):
        """returns a hubspot3 form submissions client"""
        from hubspot3.forms import FormSubmissionClient

        return FormSubmissionClient(**self.auth, **self.options)

    @property
    def forms(self):
        """returns a hubspot3 forms client"""
        from hubspot3.forms import FormsClient

        return FormsClient(**self.auth, **self.options)

    @property
    def keywords(self):
        """returns a hubspot3 keywords client"""
        from hubspot3.keywords import KeywordsClient

        return KeywordsClient(**self.auth, **self.options)

    @property
    def leads(self):
        """returns a hubspot3 leads client"""
        from hubspot3.leads import LeadsClient

        return LeadsClient(**self.auth, **self.options)

    @property
    def lines(self):
        """returns a hubspot3 lines client"""
        from hubspot3.lines import LinesClient

        return LinesClient(**self.auth, **self.options)

    @property
    def oauth2(self):
        """returns a hubspot3 OAuth2 client"""
        from hubspot3.oauth2 import OAuth2Client

        return OAuth2Client(**self.auth, **self.options)

    @property
    def owners(self):
        """returns a hubspot3 owners client"""
        from hubspot3.owners import OwnersClient

        return OwnersClient(**self.auth, **self.options)

    @property
    def products(self):
        """returns a hubspot3 products client"""
        from hubspot3.products import ProductsClient

        return ProductsClient(**self.auth, **self.options)

    @property
    def properties(self):
        """returns a hubspot3 deal properties client"""
        from hubspot3.properties import PropertiesClient

        return PropertiesClient(**self.auth, **self.options)

    @property
    def property_groups(self):
        """returns a hubspot3 property_groups client"""
        from hubspot3.property_groups import PropertyGroupsClient

        return PropertyGroupsClient(**self.auth, **self.options)

    @property
    def prospects(self):
        """returns a hubspot3 prospects client"""
        from hubspot3.prospects import ProspectsClient

        return ProspectsClient(**self.auth, **self.options)

    @property
    def settings(self):
        """returns a hubspot3 settings client"""
        from hubspot3.settings import SettingsClient

        return SettingsClient(**self.auth, **self.options)

    @property
    def tickets(self):
        """returns a hubspot3 tickets client"""
        from hubspot3.tickets import TicketsClient

        return TicketsClient(**self.auth, **self.options)

    @property
    def usage_limits(self):
        """fetches and returns a nice usage liimitation object"""
        if self._usage_limits.need_update:
            limits = self._base._call("integrations/v1/limit/daily")[0]
            self._usage_limits = Hubspot3UsageLimits(
                collected_at=datetime.fromtimestamp(int(limits["collectedAt"]) / 1000),
                current_usage=int(limits["currentUsage"]),
                fetch_status=limits["fetchStatus"],
                resets_at=datetime.fromtimestamp(int(limits["resetsAt"]) / 1000),
                usage_limit=limits["usageLimit"],
            )
        return self._usage_limits

    @property
    def me(self):
        """
        returns info about your hubspot account
        :see: https://developers.hubspot.com/docs/methods/get-account-details
        """
        return self._base._call("integrations/v1/me")



"""
hubspot products api
"""
from hubspot3.base import BaseClient
from hubspot3.utils import prettify, get_log
from typing import List


PRODUCTS_API_VERSION = "1"


class ProductsClient(BaseClient):
    """
    Products extension for products API endpoint
    THIS API ENDPOINT IS ONLY A PREVIEW AND IS SUBJECT TO CHANGE
    :see: https://developers.hubspot.com/docs/methods/products/products-overview
    """

    def __init__(self, *args, **kwargs) -> None:
        """initialize a products client"""
        super(ProductsClient, self).__init__(*args, **kwargs)
        self.log = get_log("hubspot3.products")

    def get_product(self, product_id: str, properties: List[str] = None, **options):
        """get single product based on product ID in the hubspot account"""
        properties = properties or []
        return self._call(
            "objects/products/{product_id}".format(product_id=product_id),
            method="GET",
            params={"properties": ["name", "description", *properties]},
            doseq=True,
            **options
        )

    def get_all_products(
        self, properties: List[str] = None, offset: int = 0, **options
    ):
        """get all products in the hubspot account"""
        properties = properties or []
        finished = False
        output = []
        querylimit = 100  # Max value according to docs
        while not finished:
            batch = self._call(
                "objects/products/paged",
                method="GET",
                params={
                    "limit": querylimit,
                    "offset": offset,
                    "properties": ["name", "description", *properties],
                },
                doseq=True,
                **options
            )
            output.extend(
                [
                    prettify(obj, id_key="objectId")
                    for obj in batch["objects"]
                    if not obj["isDeleted"]
                ]
            )
            finished = not batch["hasMore"]
            offset = batch["offset"]

        return output

    def _get_path(self, subpath: str):
        return "crm-objects/v{}/{}".format(
            self.options.get("version") or PRODUCTS_API_VERSION, subpath
        )

    def create(self, data=None, **options):
        """Create a new product."""
        data = data or {}

        # See: https://developers.hubspot.com/docs/methods/products/create-product
        data = [{"name": name, "value": value} for name, value in data.items()]

        return self._call("", data=data, method="POST", **options)

    def update(self, product_id, data=None, **options):
        """Update a product based on its product ID."""
        data = data or {}
        return self._call("{}".format(product_id), data=data, method="POST", **options)

    def delete(self, product_id, **options):
        """Delete a product based on its product ID."""
        return self._call("{}".format(product_id), method="DELETE", **options)
